#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Utility to handle typed lists."""

import itertools
from typing import Any, Dict, Generic, List, Optional, Set, Tuple, TypeVar, Union

from pddl.custom_types import name, parse_name, parse_type
from pddl.helpers.base import check, safe_index
from pddl.logic.functions import NumericFunction
from pddl.logic.terms import _print_tag_set
from pddl.parser.symbols import Symbols

T = TypeVar("T", name, NumericFunction)


class TypedListParser(Generic[T]):
    """
    An index for PDDL types and PDDL names/variables/functions.

    This class is used to index PDDL names, variables, and functions by their types.
    OrderedDict is used to preserve the order of the types and the names, e.g. for predicate variables.

    Other types of validations are performed to ensure that the types index is consistent.
    """

    def __init__(self, allow_duplicates: bool) -> None:
        """Initialize the types index."""
        self._allow_duplicates = allow_duplicates

        self._types_to_items: Dict[T, Set[T]] = {}
        self._item_to_types: Dict[T, Set[T]] = {}
        self._item_to_types_sequence: List[Tuple[T, Set[T]]] = []

    def add_item(self, item_name: T, type_tags: Set[T]) -> None:
        """
        Add an item to the types index with the given type tags.

        Both the item name and the type tags are validated according to the name type regular expression.

        :param item_name: the item name
        :param type_tags: the types for the item
        """
        if not self._allow_duplicates:
            self._check_item_name_already_present(item_name)
            self._check_tags_already_present(item_name, type_tags)
        self._check_item_types(item_name, type_tags)
        self._add_item(item_name, type_tags)

    def get_typed_list_of_names(self) -> Dict[T, Optional[T]]:
        """Get the typed list of names in form of dictionary."""
        result: Dict[T, Optional[T]] = {}
        for item, types_tags in self._item_to_types.items():
            if len(types_tags) > 1:
                self._raise_multiple_types_error(item, types_tags)
            type_tag = next(iter(types_tags)) if len(types_tags) == 1 else None
            result[item] = type_tag
        return result

    def get_typed_list_of_variables(self) -> Tuple[Tuple[T, Set[T]], ...]:
        """Get the typed list of variables in form of a tuple of pairs."""
        return tuple(self._item_to_types_sequence)

    @classmethod
    def parse_typed_list(
        cls, tokens: List[Union[T, List[T]]], allow_duplicates: bool = False
    ) -> "TypedListParser":
        """
        Parse typed list.

        This method takes in input a list of tokens as returned by the domain or problem parsers,
        and returns a TypesIndex object.

        The input list of tokens must have the following format:
        - if the list is not typed, it is simply a list of names
        - if the list is typed, the format is: [name_1, ..., name_n], "-", [type_1, ..., type_m], ...

        >>> index = TypedListParser.parse_typed_list(["a", "b", "c"])
        >>> index.get_typed_list_of_names()
        {'a': None, 'b': None, 'c': None}

        >>> index = TypedListParser.parse_typed_list(["a", "b", "c", "-", ["t1"]])
        >>> index.get_typed_list_of_names()
        {'a': 't1', 'b': 't1', 'c': 't1'}

        >>> index = TypedListParser.parse_typed_list(["a", "b", "c", "-", ["t1", "t2"]])
        >>> index.get_typed_list_of_names()
        Traceback (most recent call last):
        ...
        ValueError: typed list names should not have more than one type, got 'a' with types ['t1', 't2']

        :param tokens: the list of tokens
        :param allow_duplicates: whether allowed_duplicates are allowed
        :return: the TypesIndex object
        """
        result = TypedListParser(allow_duplicates=allow_duplicates)

        type_sep_index = safe_index(tokens, Symbols.TYPE_SEP.value)

        if type_sep_index is None:
            # simple list of names
            cls._add_typed_lists(result, 0, len(tokens), tokens, set())
            return result

        # if we are here, the matched pattern is: [name_1, ..., name_n], "-", parent_name, ...
        # Consume typed sublists iteratively. Caveat: the last typed list *might* have no parent type.

        # the index of the separator symbol for the current typed sublist being processed
        type_sep_index = safe_index(tokens, Symbols.TYPE_SEP.value)
        # the index of the first element of the current typed sublist
        start_index = 0
        # the index of the last element of the current typed sublist
        end_index = len(tokens)
        while type_sep_index is not None:
            # the name of the parent type is the element after the separator
            parent_names = tokens[type_sep_index + 1]
            # handle the case of a single parent type
            if not isinstance(parent_names, list):
                parent_names = [parent_names]

            # parse the typed list
            cls._add_typed_lists(
                result, start_index, type_sep_index, tokens, set(parent_names)
            )

            # go to next typed list (if any)
            start_index = type_sep_index + 2
            type_sep_index = safe_index(
                tokens, Symbols.TYPE_SEP.value, start_index, end_index
            )

        # this is to handle the case the last sublist is not typed
        if start_index != end_index:
            # parse the last typed list, with no type.
            cls._add_typed_lists(result, start_index, end_index, tokens, set())

        return result

    @classmethod
    def _add_typed_lists(
        cls,
        result: "TypedListParser",
        start_index: int,
        end_index: int,
        tokens: List[Union[T, List[T]]],
        type_tags: Set[str],
    ) -> None:
        """
        Merge typed lists.

        Side-effect on the 'result' dictionary. The start_index and end_index are needed to avoid useless
        sublist copies.
        """
        for item_name in itertools.islice(tokens, start_index, end_index):
            check(
                isinstance(item_name, str) or isinstance(item_name, NumericFunction),
                f"invalid item '{item_name}' in typed list",
            )
            # these lines implicitly perform name validation
            cast_item_name: Any = (
                parse_name(item_name) if isinstance(item_name, str) else item_name
            )
            type_tags_names: Set[Any] = set(map(parse_type, type_tags))
            result.add_item(cast_item_name, type_tags_names)

    def _check_item_name_already_present(self, item_name: T) -> None:
        """
        Check if the item name/function is already present in the index.

        :param item_name: the item name
        """
        if item_name in self._item_to_types:
            types_list = sorted(map(str, self._item_to_types[item_name]))
            if len(types_list) > 0:
                raise ValueError(
                    f"duplicate name '{item_name}' in typed list already inherits from types {types_list}"
                )
            raise ValueError(
                f"duplicate name '{item_name}' in typed list already present"
            )

    def _check_tags_already_present(self, item_name: T, type_tags: Set[T]) -> None:
        """
        Check if the type tags are already present for the given item name.

        :param item_name: the item name
        :param type_tags: the type tags
        """
        existing_tags = self._item_to_types.get(item_name, set())
        for type_tag in type_tags:
            if type_tag in existing_tags:
                raise ValueError(
                    f"duplicate type tag '{type_tag}' in typed list: type already specified for item {item_name}"
                )

    def _check_item_types(self, item_name: T, type_tags: Set[T]) -> None:
        """Check if the types of the item are valid."""
        if item_name in self._item_to_types:
            previous_type_tags = self._item_to_types[item_name]
            if previous_type_tags != type_tags:
                raise ValueError(
                    f"invalid types for item '{item_name}': previous known tags were "
                    f"{_print_tag_set(previous_type_tags)}, got {_print_tag_set(type_tags)}"
                )

    def _add_item(self, item_name: T, type_tags: Set[T]) -> None:
        """Add an item (no validation)."""
        for type_tag in type_tags:
            self._types_to_items.setdefault(type_tag, set()).add(item_name)
        self._item_to_types.setdefault(item_name, set()).update(type_tags)
        self._item_to_types_sequence.append((item_name, type_tags))

    def _raise_multiple_types_error(self, item_name: T, type_tags: Set[T]) -> None:
        """Raise an error if the item has multiple types."""
        raise ValueError(
            f"typed list names should not have more than one type, got '{item_name}' with "
            f"types {_print_tag_set(type_tags)}"
        )



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Test pddl.logic.predicates module."""

import pytest

from pddl.logic import Predicate
from pddl.logic.predicates import EqualTo
from pddl.logic.terms import Variable


def test_inconsistent_predicate_terms() -> None:
    """Test that terms of a predicate must have consistent typing."""
    with pytest.raises(
        ValueError,
        match=r"Term \?a has inconsistent type tags: previous type tags \['t1', 't2'\], new type tags "
        r"\['t3', 't4'\]",
    ):
        a1, a2 = Variable("a", ["t1", "t2"]), Variable("a", ["t3", "t4"])
        Predicate("p", a1, a2)


def test_inconsistent_equal_to_terms() -> None:
    """Test that terms of a EqualTo atomic must have consistent typing."""
    with pytest.raises(
        ValueError,
        match=r"Term \?a has inconsistent type tags: previous type tags \['t1', 't2'\], new type tags "
        r"\['t3', 't4'\]",
    ):
        a1, a2 = Variable("a", ["t1", "t2"]), Variable("a", ["t3", "t4"])
        EqualTo(a1, a2)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Helper functions."""



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains the tests for the domain parser."""

from textwrap import dedent
from typing import cast

import pytest

from pddl.exceptions import PDDLParsingError, PDDLValidationError
from pddl.logic.base import And, ExistsCondition, ForallCondition, Not
from pddl.logic.functions import (
    BinaryFunction,
    GreaterEqualThan,
    Increase,
    NumericFunction,
    NumericValue,
)
from pddl.logic.predicates import EqualTo, Predicate
from pddl.logic.terms import Variable
from pddl.parser.domain import DomainParser
from pddl.parser.symbols import Symbols
from tests.conftest import TEXT_SYMBOLS


def test_hierarchical_types() -> None:
    """Test correct parsing of hierarchical types (see https://github.com/AI-Planning/pddl/issues/70)."""
    domain_str = dedent("""
    (define (domain logistics)
        (:requirements :strips :typing)
        (:types truck airplane - vehicle
            package vehicle - physobj
            airport location - place
            city place physobj - object)
        (:predicates (in-city ?loc - place ?city - city)
            (at ?obj - physobj ?loc - place)
            (in ?pkg - package ?veh - vehicle))
        (:action LOAD-TRUCK
            :parameters   (?pkg - package ?truck - truck ?loc - place)
            :precondition (and (at ?truck ?loc) (at ?pkg ?loc))
            :effect       (and (not (at ?pkg ?loc)) (in ?pkg ?truck)))
    )
    """)
    domain = DomainParser()(domain_str)

    assert domain.types == {
        "truck": "vehicle",
        "airplane": "vehicle",
        "package": "physobj",
        "vehicle": "physobj",
        "airport": "place",
        "location": "place",
        "city": None,
        "place": None,
        "physobj": None,
    }


def test_hierarchical_types_2() -> None:
    """Test correct parsing of hierarchical types, Storage domain."""
    domain_str = dedent("""
    (define (domain logistics)
        (:requirements :strips :typing)
        (:types hoist surface place area - object
            container depot - place
            storearea transitarea - area
            crate - surface)
    )
    """)
    domain = DomainParser()(domain_str)
    assert domain.types == {
        "hoist": None,
        "surface": None,
        "place": None,
        "area": None,
        "container": "place",
        "depot": "place",
        "storearea": "area",
        "transitarea": "area",
        "crate": "surface",
    }


def test_types_repetition_in_simple_typed_lists_not_allowed() -> None:
    """Check types repetition in simple typed lists is detected and a parsing error is raised."""
    domain_str = dedent("""
    (define (domain test)
        (:requirements :typing)
        (:types a b c a)
    )
    """)

    with pytest.raises(
        PDDLParsingError,
        match=r".*error while parsing tokens \['a', 'b', 'c', 'a'\]: "
        "duplicate name 'a' in typed list already present",
    ):
        DomainParser()(domain_str)


def test_types_repetition_in_typed_lists_not_allowed() -> None:
    """Check types repetition in typed lists is detected and a parsing error is raised."""
    domain_str = dedent("""
    (define (domain test)
        (:requirements :typing)
        (:types a - t1 b c - t2 a - t3)
    )
    """)

    with pytest.raises(
        PDDLParsingError,
        match=r".*error while parsing tokens \['a', '-', 't1', 'b', 'c', '-', 't2', 'a', '-', 't3'\]: "
        r"duplicate name 'a' in typed list already inherits from types \['t1'\]",
    ):
        DomainParser()(domain_str)


def test_typing_requirement_under_other_domain_requirements() -> None:
    """Check :typing requirement does not throw error if other domain requirements that includes it are detected."""
    domain_str = dedent("""
(define (domain test)
  (:requirements :adl)
  (:types a b c)
  (:predicates
    (predicate1 ?x - a)
    (predicate2 ?x - b)
    (predicate3 ?x - c)
    )
  )
    """)

    domain = DomainParser()(domain_str)
    assert domain.types == {
        "a": None,
        "b": None,
        "c": None,
    }


@pytest.mark.parametrize("keyword", TEXT_SYMBOLS - {Symbols.OBJECT.value})
def test_keyword_usage_not_allowed_as_name(keyword) -> None:
    """Check keywords usage as names is detected and a parsing error is raised."""
    domain_str = dedent(f"""
    (define (domain test)
        (:requirements :typing)
        (:types {keyword})
    )
    """)

    with pytest.raises(
        PDDLValidationError,
        match=f".*invalid name '{keyword}': it is a keyword",
    ):
        DomainParser()(domain_str)


@pytest.mark.parametrize("keyword", TEXT_SYMBOLS - {Symbols.OBJECT.value})
def test_keyword_usage_not_allowed_as_type(keyword) -> None:
    """Check keywords usage as types is detected and a parsing error is raised."""
    domain_str = dedent(f"""
    (define (domain test)
        (:requirements :typing)
        (:types t1 - {keyword})
    )
    """)

    with pytest.raises(
        PDDLValidationError,
        match=f".*invalid type '{keyword}': it is a keyword",
    ):
        DomainParser()(domain_str)


def test_constants_repetition_in_simple_typed_lists_not_allowed() -> None:
    """Check constants repetition in simple typed lists is detected and a parsing error is raised."""
    domain_str = dedent("""
    (define (domain test)
        (:requirements :typing)
        (:types t1)
        (:constants c1 c2 c3 c1)
    )
    """)

    with pytest.raises(
        PDDLParsingError,
        match=".*error while parsing tokens \\['c1', 'c2', 'c3', 'c1'\\]: "
        "duplicate name 'c1' in typed list already present",
    ):
        DomainParser()(domain_str)


def test_constants_repetition_in_typed_lists_not_allowed() -> None:
    """Check constants repetition in typed lists is detected and a parsing error is raised."""
    domain_str = dedent("""
    (define (domain test)
        (:requirements :typing)
        (:types t1 t2)
        (:constants c1 - t1 c1 - t2)
    )
    """)

    with pytest.raises(
        PDDLParsingError,
        match=".*error while parsing tokens \\['c1', '-', 't1', 'c1', '-', 't2'\\]: "
        "duplicate name 'c1' in typed list already inherits from types \\['t1'\\]",
    ):
        DomainParser()(domain_str)


def test_variables_repetition_in_simple_typed_lists_allowed() -> None:
    """Check variables repetition in simple typed lists is allowed."""
    domain_str = dedent("""
    (define (domain test)
        (:requirements :typing)
        (:predicates (p ?x ?y ?z ?x))
    )
    """)

    DomainParser()(domain_str)


def test_variables_repetition_in_typed_lists_not_allowed() -> None:
    """Check variables repetition in typed lists is detected and a parsing error is raised."""
    domain_str = dedent("""
    (define (domain test)
        (:requirements :typing)
        (:types t1 t2)
        (:predicates (p ?x - (either t1 t2) ?x - t3))
    )
    """)

    with pytest.raises(
        PDDLParsingError,
        match=r".*error while parsing tokens \['x', '-', \['t1', 't2'\], 'x', '-', \['t3'\]\]: "
        r"invalid types for item \'x\': previous known tags were \['t1', 't2'\], got \['t3'\]",
    ):
        DomainParser()(domain_str)


def test_variables_typed_with_not_available_types() -> None:
    """Check variables with non-available types raises a parsing error."""
    domain_str = dedent("""
    (define (domain test)
        (:requirements :typing)
        (:types t1)
        (:predicates (p ?x - t2))
    )
    """)

    with pytest.raises(
        PDDLValidationError,
        match=r"types \['t2'\] of term Variable\(x\) are not in available types \{'t1'\}",
    ):
        DomainParser()(domain_str)


def test_variables_repetition_allowed_if_same_type() -> None:
    """Check variables repetition in typed lists is allowed."""
    domain_str = dedent("""
    (define (domain test)
        (:requirements :typing :existential-preconditions :universal-preconditions)
        (:types t1 t2)
        (:predicates (p ?x - (either t1 t2) ?x - (either t1 t2)))
        (:action a
            :parameters (?x - t1 ?x - t1)
            :precondition (and (exists (?x - t1) (p ?x ?x)) (forall (?x - t1) (p ?x ?x)))
            :effect (p ?x ?x)
        )
    )
    """)
    DomainParser()(domain_str)


def test_action_parameter_type_propagation_in_precondition_and_effect() -> None:
    """Test that the type tags of vars in single action parameters propagate correctly in precondition/effect terms."""
    # single param ?v - mytype; ensure propagation to precondition and effect
    domain_str = dedent("""
    (define (domain action_type_propagation)
        (:requirements :typing)
        (:types mytype - object)
        (:predicates
           (P ?x - mytype)
           (Q ?x)                ; untyped here on purpose
        )
        (:action use_v
          :parameters (?v - mytype)
          :precondition (Q ?v)
          :effect (P ?v))
    )
    """)
    domain = DomainParser()(domain_str)

    action = next(iter(domain.actions))
    assert action.name == "use_v"

    param_v = action.parameters[0]
    assert param_v.type_tags == {"mytype"}

    assert type(action.precondition) is Predicate
    v_in_pre = action.precondition.terms[0]

    assert type(action.effect) is Predicate
    v_in_eff = action.effect.terms[0]

    # type should be propagated to both occurrences within the action's scope
    assert v_in_pre.type_tags == {"mytype"}
    assert v_in_eff.type_tags == {"mytype"}

    # and the variables should be "the same variable" within the same scope
    assert param_v == v_in_pre
    assert param_v == v_in_eff
    assert v_in_pre == v_in_eff


def test_action_multiple_parameters_each_propagates_independently() -> None:
    """Test that the type tags of vars in many actions parameters propagate correctly in precondition/effect terms."""
    # two params ?x - mytype1 and ?y - mytype2; ensure each keeps its own types
    domain_str = dedent("""
    (define (domain action_two_params)
        (:requirements :typing)
        (:types mytype1 mytype2)
        (:predicates
           (R ?x - mytype1 ?y - mytype2)
           (Q ?x)  ; untyped, to check propagation from parameters
           (S ?y)  ; untyped, to check propagation from parameters
        )
        (:action use_xy
          :parameters (?x - mytype1 ?y - mytype2)
          :precondition (Q ?x)
          :effect (R ?x ?y))
        (:action use_yx
          :parameters (?x - mytype1 ?y - mytype2)
          :precondition (S ?y)
          :effect (R ?x ?y))
    )
    """)
    domain = DomainParser()(domain_str)

    # ---- Check variable types in predicate *definitions* ----
    preds = {p.name: p for p in domain.predicates}
    pred_R = preds["R"]
    pred_Q = preds["Q"]
    pred_S = preds["S"]

    x_in_R_def = pred_R.terms[0]
    y_in_R_def = pred_R.terms[1]
    assert x_in_R_def.type_tags == {"mytype1"}
    assert y_in_R_def.type_tags == {"mytype2"}

    x_in_Q_def = pred_Q.terms[0]
    y_in_S_def = pred_S.terms[0]
    assert x_in_Q_def.type_tags == set()
    assert y_in_S_def.type_tags == set()
    # ---------------------------------------------------------

    # Sort for deterministic order
    actions = sorted(domain.actions, key=lambda a: a.name)
    action1 = actions[0]  # use_xy
    action2 = actions[1]  # use_yx
    assert action1.name == "use_xy"
    assert action2.name == "use_yx"

    # ---------- Action 1: use_xy ----------
    x_param1, y_param1 = action1.parameters
    assert x_param1.type_tags == {"mytype1"}
    assert y_param1.type_tags == {"mytype2"}

    # Precondition uses only ?x
    assert type(action1.precondition) is Predicate
    x_in_pre1 = action1.precondition.terms[0]
    assert x_in_pre1.type_tags == {"mytype1"}
    assert x_in_pre1 == x_param1

    # Effect uses both ?x and ?y, in order
    assert type(action1.effect) is Predicate
    x_in_eff1 = action1.effect.terms[0]
    y_in_eff1 = action1.effect.terms[1]
    assert x_in_eff1.type_tags == {"mytype1"}
    assert y_in_eff1.type_tags == {"mytype2"}

    # Identity preserved per variable
    assert x_in_eff1 == x_param1
    assert y_in_eff1 == y_param1
    assert x_in_eff1 != y_in_eff1

    # ---------- Action 2: use_yx ----------
    x_param2, y_param2 = action2.parameters
    assert x_param2.type_tags == {"mytype1"}
    assert y_param2.type_tags == {"mytype2"}

    # Precondition uses only ?y
    assert type(action2.precondition) is Predicate
    y_in_pre2 = action2.precondition.terms[0]
    assert y_in_pre2.type_tags == {"mytype2"}
    assert y_in_pre2 == y_param2

    # Effect uses both ?x and ?y, in order
    assert type(action2.effect) is Predicate
    x_in_eff2 = action2.effect.terms[0]
    y_in_eff2 = action2.effect.terms[1]
    assert x_in_eff2.type_tags == {"mytype1"}
    assert y_in_eff2.type_tags == {"mytype2"}

    # Identity preserved per variable
    assert x_in_eff2 == x_param2
    assert y_in_eff2 == y_param2
    assert x_in_eff2 != y_in_eff2


def test_variables_types_propagated_in_derived_predicate() -> None:
    """Test that variables occurring in definition of derived predicate propagated in its condition."""
    domain_str = dedent("""
    (define (domain samevariabledifferent)
        (:requirements :typing)
        (:types mytype)
        (:predicates
           (P ?v - mytype)
           (Q ?v)
        )
        (:derived (P ?v)
          (Q ?v))
        )
    """)
    domain = DomainParser()(domain_str)

    # check that type of variable ?v in predicates is parsed correctly
    _predicates_sorted_by_name = sorted(domain.predicates, key=lambda p: p.name)
    predicate_p = _predicates_sorted_by_name[0]
    predicate_q = _predicates_sorted_by_name[1]
    var_v_in_pred_p = predicate_p.terms[0]
    var_v_in_pred_q = predicate_q.terms[0]
    assert var_v_in_pred_p.type_tags == {"mytype"}
    assert var_v_in_pred_q.type_tags == set()
    assert var_v_in_pred_p != var_v_in_pred_q

    # check that type of variable ?v in derived predicate is propagated even if type does not occur
    axiom = next(iter(domain.derived_predicates))
    assert type(axiom.predicate) is Predicate
    var1 = axiom.predicate.terms[0]
    assert type(axiom.condition) is Predicate
    var2 = axiom.condition.terms[0]
    assert var1.type_tags == var2.type_tags == {"mytype"}
    assert var1 == var2


def test_variables_types_propagated_in_derived_predicate_complex_condition() -> None:
    """Test that variables occurring in definition of derived predicate propagated in its complex condition."""
    domain_str = dedent("""
    (define (domain samevariabledifferent)
        (:requirements :typing :equality :quantified-preconditions)
        (:types mytype mytype2 mytype3)
        (:predicates
           (P ?v - mytype)
           (Q ?v)
           (R ?v)
        )
        (:derived (P ?v)
          (and
            (Q ?v)
            (not (R ?v))
            (exists (?v - mytype2) (Q ?v))
            (forall (?v - mytype3) (Q ?v))
            (= ?v ?v)
            (>= 1 0)
          )
        )
    )
    """)
    domain = DomainParser()(domain_str)

    # check that type of variable ?v in predicates is parsed correctly
    mytype = "mytype"
    mytype2 = "mytype2"
    mytype3 = "mytype3"
    _predicates_sorted_by_name = sorted(domain.predicates, key=lambda p: p.name)
    predicate_p = _predicates_sorted_by_name[0]
    predicate_q = _predicates_sorted_by_name[1]
    predicate_r = _predicates_sorted_by_name[2]
    var_v_in_pred_p = predicate_p.terms[0]
    var_v_in_pred_q = predicate_q.terms[0]
    var_v_in_pred_r = predicate_r.terms[0]
    assert var_v_in_pred_p.type_tags == {mytype}
    assert var_v_in_pred_q.type_tags == set()
    assert var_v_in_pred_r.type_tags == set()
    assert var_v_in_pred_p != var_v_in_pred_q

    # check that type of variable ?v in derived predicate is propagated even if type does not occur
    axiom = next(iter(domain.derived_predicates))
    assert type(axiom.predicate) is Predicate
    var1 = axiom.predicate.terms[0]

    assert type(axiom.condition) is And
    operands = axiom.condition.operands
    pred_q, not_pred_r, exists_pred_v2, forall_pred_v3, eq_v_v, ge_pred_1_0 = operands

    assert type(pred_q) is Predicate
    q_var = pred_q.terms[0]
    assert q_var.type_tags == {mytype}
    assert var1 == q_var

    assert type(not_pred_r) is Not
    assert type(not_pred_r.argument) is Predicate
    not_var = not_pred_r.argument.terms[0]
    assert not_var.type_tags == {mytype}
    assert var1 == not_var

    assert type(exists_pred_v2) is ExistsCondition
    assert list(exists_pred_v2.variables)[0].type_tags == {mytype2}
    assert type(exists_pred_v2.condition) is Predicate
    exists_var = exists_pred_v2.condition.terms[0]
    assert exists_var.type_tags == {mytype2}
    assert var1 != exists_var

    assert type(forall_pred_v3) is ForallCondition
    assert list(forall_pred_v3.variables)[0].type_tags == {mytype3}
    assert type(forall_pred_v3.condition) is Predicate
    forall_var = forall_pred_v3.condition.terms[0]
    assert forall_var.type_tags == {mytype3}
    assert var1 != forall_var

    assert type(eq_v_v) is EqualTo
    eq_v1 = eq_v_v.left
    eq_v2 = eq_v_v.right
    assert eq_v1.type_tags == {mytype}
    assert var1 == eq_v1
    assert eq_v2.type_tags == {mytype}
    assert var1 == eq_v2

    assert type(ge_pred_1_0) is GreaterEqualThan


def test_variables_types_propagated_in_derived_predicate_complex_hierarchy() -> None:
    """Test that variables occurring in definition of derived predicate propagated in its condition."""
    domain_str = dedent("""
    (define (domain samevariabledifferent)
        (:requirements :typing)
        (:types child1 - root_type
                child2 - child1
                child3 - root_type)
        (:predicates
           (P ?v - root_type)
           (Q ?v)
           (R ?v)
        )
        (:derived (P ?v - child1) (Q ?v))
        (:derived (P ?v - child3) (R ?v))
    )
    """)
    domain = DomainParser()(domain_str)
    axioms = sorted(
        domain.derived_predicates, key=lambda dp: cast(Predicate, dp.condition).name
    )
    axiom_pq = axioms[0]
    axiom_pr = axioms[1]

    assert axiom_pq.predicate.terms[0].type_tags == {"child1"}
    assert cast(Predicate, axiom_pq.condition).terms[0].type_tags == {"child1"}

    assert axiom_pr.predicate.terms[0].type_tags == {"child3"}
    assert cast(Predicate, axiom_pr.condition).terms[0].type_tags == {"child3"}


def test_check_action_costs_requirement_with_total_cost() -> None:
    """Check action costs requirement when total-cost is specified."""
    domain_str = dedent("""
    (define (domain test)
        (:requirements :typing)
        (:types t1 t2)
        (:predicates (p ?x - t1 ?y - t2))
        (:functions (total-cost))
        (:action a
            :parameters (?x - t1 ?y - t2)
            :precondition (and (p ?x ?x))
            :effect (and (p ?x ?x) (increase (total-cost) 1))
        )
    )
    """)
    with pytest.raises(
        PDDLValidationError,
        match=r"action costs requirement is not specified, but the total-cost function is specified.",
    ):
        DomainParser()(domain_str)


def test_variable_types_in_strips_action_definition() -> None:
    """Check typing for predicate variables in action preconditions and effects."""
    domain_str = dedent("""
    (define (domain test)
        (:requirements :typing)
        (:types t1 t2)
        (:predicates (p ?x - t1 ?y - t2))
        (:action a
            :parameters (?x - t1 ?y - t2)
            :precondition (p ?x ?y)
            :effect (p ?x ?y)
        )
    )
    """)
    domain = DomainParser()(domain_str)
    action = next(iter(domain.actions))
    x = Variable("x", {"t1"})
    y = Variable("y", {"t2"})
    assert action.parameters == (x, y)
    assert isinstance(action.precondition, Predicate)
    assert action.precondition.terms == (x, y)
    assert isinstance(action.effect, Predicate)
    assert action.effect.terms == (x, y)


def test_variable_types_in_numeric_action_definition() -> None:
    """Check typing for function variables in action preconditions and effects."""
    domain_str = dedent("""
    (define (domain test)
        (:requirements :typing :numeric-fluents)
        (:types t1 t2)
        (:functions (f ?x - t1 ?y - t2))
        (:action a
            :parameters (?x - t1 ?y - t2)
            :precondition (<= 1 (f ?x ?y))
            :effect (increase (f ?x ?y) 1)
        )
    )
    """)
    domain = DomainParser()(domain_str)
    action = next(iter(domain.actions))
    x = Variable("x", {"t1"})
    y = Variable("y", {"t2"})
    assert action.parameters == (x, y)
    assert isinstance(action.precondition, BinaryFunction)
    assert isinstance(action.precondition.operands[1], NumericFunction)
    assert action.precondition.operands[1].terms == (x, y)
    assert isinstance(action.effect, Increase)
    assert isinstance(action.effect.operands[0], NumericFunction)
    assert action.effect.operands[0].terms == (x, y)


def test_number_parsing() -> None:
    """Check parsing of numbers in  action preconditions and effects."""
    domain_str = dedent("""
    (define (domain test)
        (:requirements :typing :numeric-fluents)
        (:types t1 t2)
        (:functions (f ?x - t1 ?y - t2) (g ?x - t2))
        (:action a
            :parameters (?x - t1 ?y - t2)
            :precondition (and (<= 10 (f ?x ?y)) (> (g ?y) 2.5))
            :effect (and (decrease (f ?x ?y) 42) (increase (g ?y) 0.5))
        )
    )
    """)
    domain = DomainParser()(domain_str)
    action = next(iter(domain.actions))
    x = Variable("x", {"t1"})
    y = Variable("y", {"t2"})
    assert action.parameters == (x, y)
    assert isinstance(action.precondition, And)
    f_precond = action.precondition.operands[0]
    assert isinstance(f_precond, BinaryFunction)
    g_precond = action.precondition.operands[1]
    assert isinstance(g_precond, BinaryFunction)
    assert type(f_precond.operands[0]) is NumericValue
    assert type(f_precond.operands[0].value) is int
    assert f_precond.operands[0] == NumericValue(10)
    assert isinstance(f_precond.operands[1], NumericFunction)
    assert f_precond.operands[1].terms == (x, y)
    assert type(g_precond.operands[1]) is NumericValue
    assert type(g_precond.operands[1].value) is float
    assert g_precond.operands[1] == NumericValue(2.5)
    assert isinstance(g_precond.operands[0], NumericFunction)
    assert g_precond.operands[0].terms == (y,)
    assert isinstance(action.effect, And)
    f_effect = action.effect.operands[0]
    assert isinstance(f_effect, BinaryFunction)
    assert isinstance(f_effect.operands[0], NumericFunction)
    assert isinstance(f_effect.operands[1], NumericValue)
    g_effect = action.effect.operands[1]
    assert isinstance(g_effect, BinaryFunction)
    assert isinstance(g_effect.operands[0], NumericFunction)
    assert isinstance(g_effect.operands[1], NumericValue)
    assert f_effect.operands[0].terms == (x, y)
    assert type(f_effect.operands[1]) is NumericValue
    assert type(f_effect.operands[1].value) is int
    assert f_effect.operands[1] == NumericValue(42)
    assert g_effect.operands[0].terms == (y,)
    assert type(g_effect.operands[1]) is NumericValue
    assert type(g_effect.operands[1].value) is float
    assert g_effect.operands[1] == NumericValue(0.5)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This test module contains the fixtures for 'blocksworld-ipc08' domain and problem."""

import pytest

from pddl.action import Action
from pddl.core import Domain, Problem
from pddl.logic import Constant
from pddl.logic.base import And, OneOf
from pddl.logic.effects import When
from pddl.logic.helpers import constants, variables
from pddl.logic.predicates import EqualTo, Predicate
from pddl.requirements import Requirements


@pytest.fixture(scope="session")
def blocksworld_fond_domain():
    """The 'blocksworld' FOND domain."""
    # terms
    x, y, z, b = variables("x y z b")

    # constants:
    table = Constant("Table")

    # predicates
    on = Predicate("on", x, y)
    clear = Predicate("clear", x)
    block = Predicate("block", b)
    predicates = {on, clear, block}

    # actions
    # put-on
    put_on_name = "puton"
    put_on_parameters = [x, y, z]
    put_on_precondition = (
        on(x, z)
        & clear(x)
        & clear(y)
        & ~EqualTo(y, z)
        & ~EqualTo(x, z)
        & ~EqualTo(x, y)
        & ~EqualTo(x, table)
    )
    put_on_effect = OneOf(
        And(
            on(x, y),
            ~on(x, z),
            When(~EqualTo(z, table), clear(z)),
            When(~EqualTo(y, table), ~clear(y)),
        ),
        And(
            on(x, table),
            When(~EqualTo(z, table), And(~on(x, z), clear(z))),
            When(~EqualTo(y, table), ~clear(y)),
        ),
    )
    put_on = Action(put_on_name, put_on_parameters, put_on_precondition, put_on_effect)

    name = "blocks-world-domain"
    requirements = {
        Requirements.STRIPS,
        Requirements.EQUALITY,
        Requirements.NON_DETERMINISTIC,
        Requirements.CONDITIONAL_EFFECTS,
    }
    actions = {put_on}
    domain = Domain(
        name=name,
        requirements=requirements,
        constants={table},
        predicates=predicates,
        actions=actions,
    )
    return domain


@pytest.fixture(scope="session")
def blocksworld_fond_01():
    """Blocksworld FOND problem 01."""
    # objects
    objects = [A, B, C] = constants("A B C")

    Table = Constant("Table")

    # predicates
    block = Predicate("block", A)
    on = Predicate("on", C, A)
    clear = Predicate("clear", B)

    init = {
        block(A),
        block(B),
        block(C),
        block(Table),
        on(C, A),
        on(A, Table),
        on(B, Table),
        clear(C),
        clear(B),
        clear(Table),
    }

    goal = on(B, A)

    problem_name = "sussman-anomaly"

    problem = Problem(
        problem_name,
        domain_name="blocks-world-domain",
        objects=objects,
        init=init,
        goal=goal,
    )
    return problem



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Base classes for pylogics logic formulas."""

from functools import wraps
from typing import Any, Callable, cast


def _cache_hash(fn) -> Callable[[Any], int]:
    """
    Compute the (possibly memoized) hash.

    If the hash for this object has already
    been computed, return it. Otherwise,
    compute it and store for later calls.

    :param fn: the hashing function.
    :return: the new hashing function.
    """

    @wraps(fn)
    def __hash__(self):
        if not hasattr(self, "__hash"):
            self.__hash = fn(self)
        return cast(int, self.__hash)

    return __hash__


def _getstate(fn):
    """
    Get the state.

    We need to ignore the hash value because in case the object
    is serialized with e.g. Pickle, if the state is restored
    with another instance of the interpreter, the stored hash might
    be inconsistent with the PYTHONHASHSEED initialization of the
    new interpreter.

    :param fn: the getstate function.
    :return: the new getstate function.
    """

    @wraps(fn)
    def __getstate__(self):
        d = fn(self)
        if d is None:
            return None
        # assuming d is a dictionary
        d.pop("__hash", None)
        return d

    return __getstate__


def default_getstate(self):
    """Implement the default getstate."""
    return self.__dict__


def default_setstate(self, state):
    """Implement the default getstate."""
    self.__dict__ = state


def _setstate(fn):
    """
    Set the state.

    The hash value needs to be set to None
    as the state might be restored in another
    interpreter in which the old hash value
    might not be consistent anymore.

    :param fn: the setstate function.
    :return: the new setstate function.
    """

    @wraps(fn)
    def __setstate__(self, state):
        fn(self, state)
        if hasattr(self, "__hash"):
            del self.__hash

    return __setstate__


def cache_hash(cls):
    """
    Make instances of a class to cache their hash.

    This class decorator sets:
        __hash__
        __getstate__
        __setstate__

    :param cls: the class to wrap
    :return: the wrapped class
    """
    cls.__hash__ = _cache_hash(cls.__hash__)

    getstate_fn = cls.__getstate__ if hasattr(cls, "__getstate__") else default_getstate
    cls.__getstate__ = _getstate(getstate_fn)

    setstate_fn = cls.__setstate__ if hasattr(cls, "__setstate__") else default_setstate
    cls.__setstate__ = _setstate(setstate_fn)
    return cls



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains tests for action instantiation."""

import os

import pytest

from pddl import parse_domain, parse_plan, parse_problem
from pddl.action import Action
from pddl.logic.base import ExistsCondition
from pddl.logic.functions import Metric, NumericFunction, Times
from pddl.logic.helpers import constants, variables
from pddl.logic.predicates import DerivedPredicate, EqualTo, Predicate
from tests.conftest import PLAN_FILES


@pytest.mark.parametrize("plan_file", PLAN_FILES)
def test_parse_instantiate_plans(plan_file):
    """Test parsing and instantiating plans from PDDL files."""
    domain_file = os.path.dirname(plan_file) + "/domain.pddl"
    problem_name = os.path.basename(plan_file).split(".")[0]
    problem_file = os.path.dirname(plan_file) + f"/{problem_name}.pddl"

    domain = parse_domain(domain_file)
    problem = parse_problem(problem_file)
    plan = parse_plan(plan_file)

    assert domain
    assert problem
    assert plan

    for action in plan.instantiate(domain):
        assert isinstance(action, Action)
        assert "?" not in str(action)


def test_partially_instantiate_predicate():
    """Test partially instantiating a predicate."""
    a, b = constants("a b")
    x, y, z = variables("x y z")
    p = Predicate("on", x, y, z)
    mapping = {x: a, z: b}
    p_instantiated = p.instantiate(mapping)

    p_expected = Predicate("on", a, y, b)
    assert p_instantiated == p_expected


def test_instantiate_partially_instantiated_predicate():
    """Test instantiating a partially instantiated predicate."""
    a, b = constants("a b")
    x, y = variables("x y")
    p = Predicate("on", x, y)
    mapping1 = {x: a}
    p_partially_instantiated = p.instantiate(mapping1)

    mapping2 = {y: b}
    p_fully_instantiated = p_partially_instantiated.instantiate(mapping2)

    p_expected = Predicate("on", a, b)
    assert p_fully_instantiated == p_expected


def test_partially_instantiate_function():
    """Test partially instantiating a function."""
    a, b = constants("a b")
    x, y, z = variables("x y z")
    f = NumericFunction("distance", x, y, z)
    mapping = {x: a}
    f_instantiated = f.instantiate(mapping)

    f_expected = NumericFunction("distance", a, y, z)
    assert f_instantiated == f_expected


def test_instantiate_partially_instantiated_function():
    """Test instantiating a partially instantiated function."""
    a, b = constants("a b")
    x, y = variables("x y")
    f = NumericFunction("distance", x, y)
    mapping1 = {x: a}
    f_partially_instantiated = f.instantiate(mapping1)

    mapping2 = {y: b}
    f_fully_instantiated = f_partially_instantiated.instantiate(mapping2)

    f_expected = NumericFunction("distance", a, b)
    assert f_fully_instantiated == f_expected


def test_instantiate_equalto_predicate():
    """Test instantiating an EqualTo predicate."""
    x, y = variables("x y")
    a, b = constants("a b")
    eq = EqualTo(x, y)
    mapping = {x: a, y: b}
    eq_instantiated = eq.instantiate(mapping)

    eq_expected = EqualTo(a, b)
    assert eq_instantiated == eq_expected


def test_partially_instantiate_equalto_predicate():
    """Test instantiating an EqualTo predicate."""
    x, y = variables("x y")
    a, b = constants("a b")
    eq = EqualTo(x, y)
    mapping = {x: a}
    eq_instantiated = eq.instantiate(mapping)

    eq_expected = EqualTo(a, y)
    assert eq_instantiated == eq_expected

    mapping2 = {y: b}
    eq_fully_instantiated = eq_instantiated.instantiate(mapping2)

    eq_expected = EqualTo(a, b)
    assert eq_fully_instantiated == eq_expected


def test_instantiate_derived_predicate():
    """Test instantiating a derived predicate."""
    a, b = constants("a b")
    x, y, z = variables("x y z")

    p = Predicate("above", x, y)
    condition = ExistsCondition(
        variables={z}, cond=Predicate("above", x, z) & Predicate("above", z, y)
    )
    dp = DerivedPredicate(predicate=p, condition=condition)

    mapping = {x: a, y: b}
    dp_instantiated = dp.instantiate(mapping)

    p_expected = Predicate("above", a, b)
    condition_expected = ExistsCondition(
        variables={z}, cond=Predicate("above", a, z) & Predicate("above", z, b)
    )
    dp_expected = DerivedPredicate(predicate=p_expected, condition=condition_expected)

    assert dp_instantiated == dp_expected


def test_bad_action_instantiation_parameters():
    """Test action instantiation error handling."""
    a, b, c = constants("a b c")
    x, y = variables("x y")
    action = Action(
        name="move",
        parameters=[x, y],
        precondition=Predicate("at", x),
        effect=Predicate("at", y),
    )

    with pytest.raises(ValueError):
        action.instantiate([a])  # Missing one parameter

    with pytest.raises(ValueError):
        action.instantiate([a, b, c])  # One extra parameter


def test_instantiate_metric():
    """Test instantiating a metric."""
    a, b, c = constants("a b c")
    expression = Times(NumericFunction("distance", a, b), NumericFunction("cost", c))
    metric = Metric(expression)

    mapping = {}
    metric_instantiated = metric.instantiate(mapping)

    assert metric_instantiated == metric



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This class implements PDDL functions."""

import functools
from abc import abstractmethod
from collections.abc import Mapping
from typing import Sequence, cast

from pddl.custom_types import namelike, parse_function
from pddl.helpers.base import assert_
from pddl.helpers.cache_hash import cache_hash
from pddl.logic.base import Atomic, BinaryOpMetaclass, Formula
from pddl.logic.terms import Term, Variable
from pddl.parser.symbols import Symbols


@cache_hash
class FunctionExpression(Atomic):
    """A class for all the function expressions."""

    @abstractmethod
    def instantiate(self, mapping: Mapping[Variable, Term]) -> Formula:
        """Instantiate the formula with a mapping from variables to terms."""
        raise NotImplementedError()


@cache_hash
@functools.total_ordering
class NumericFunction(FunctionExpression):
    """A class for a numeric function."""

    def __init__(self, name: namelike, *terms: Term):
        """Initialize the function."""
        self._name = parse_function(name)
        self._terms = tuple(terms)

    @property
    def name(self) -> str:
        """Get the name."""
        return self._name

    @property
    def terms(self) -> Sequence[Term]:
        """Get the terms."""
        return self._terms

    @property
    def arity(self) -> int:
        """Get the arity of the function."""
        return len(self.terms)

    def instantiate(self, mapping: Mapping[Variable, Term]) -> "NumericFunction":
        """Instantiate the function with a mapping from variables to terms."""
        instantiated_terms = []
        for term in self.terms:
            if isinstance(term, Variable) and term in mapping:
                # Instantiate the variable
                instantiated_terms.append(mapping[term])
            else:
                # The function is already partially instantiated or mapping is a partial instantiation
                instantiated_terms.append(term)
        return NumericFunction(self.name, *instantiated_terms)

    def __call__(self, *terms: Term):
        """Replace terms."""
        assert_(len(terms) == self.arity, "Wrong number of terms.")
        assert_(
            all(t1.type_tags == t2.type_tags for t1, t2 in zip(self.terms, terms)),
            "Wrong types of replacements.",
        )
        return NumericFunction(self.name, *terms)

    def __str__(self) -> str:
        """Get the string."""
        if self.arity == 0:
            return f"({self.name})"
        else:
            return f"({self.name} {' '.join(map(str, self.terms))})"

    def __repr__(self) -> str:
        """Get an unambiguous string representation."""
        return f"{type(self).__name__}({self.name}, {', '.join(map(str, self.terms))})"

    def __eq__(self, other):
        """Override equal operator."""
        return (
            isinstance(other, NumericFunction)
            and self.name == other.name
            and self.terms == other.terms
        )

    def __hash__(self) -> int:
        """Compute the hash of the object."""
        return hash((self.name, self.terms))

    def __lt__(self, other):
        """Override less than operator."""
        if not isinstance(other, NumericFunction):
            return NotImplemented
        return (self.name, self.terms) < (other.name, other.terms)


@cache_hash
@functools.total_ordering
class NumericValue(FunctionExpression):
    """A class for a numeric value."""

    def __init__(self, value: float) -> None:
        """Init the numeric value object."""
        self._value = value

    @property
    def value(self) -> float:
        """Get the value."""
        return self._value

    def instantiate(self, mapping: Mapping[Variable, Term]) -> "NumericValue":
        """Instantiate the value with a mapping from variables to terms."""
        return self

    def __eq__(self, other):
        """Compare with another object."""
        return isinstance(other, NumericValue) and self.value == other.value

    def __lt__(self, other):
        """Compare with another object."""
        if not isinstance(other, NumericValue):
            return NotImplemented
        return self.value < other.value

    def __hash__(self) -> int:
        """Compute the hash of the object."""
        return hash(self._value)

    def __str__(self) -> str:
        """Get the string representation."""
        return str(self._value)


class BinaryFunction(FunctionExpression):
    """A class for a numeric binary function."""

    SYMBOL: Symbols

    def __init__(self, *operands: FunctionExpression):
        """
        Init a binary operator.

        :param operands: the operands.
        """
        self._operands = operands

    @property
    def operands(self) -> Sequence[FunctionExpression]:
        """Get the operands."""
        return tuple(self._operands)

    def instantiate(self, mapping: Mapping[Variable, Term]) -> "FunctionExpression":
        """Instantiate the formula with a mapping from variables to terms."""
        return type(self)(
            *(cast(FunctionExpression, op.instantiate(mapping)) for op in self.operands)
        )

    def __str__(self) -> str:
        """Get the string representation."""
        return f"({self.SYMBOL.value} {' '.join(map(str, self.operands))})"

    def __repr__(self) -> str:
        """Get an unambiguous string representation."""
        return f"{type(self).__name__}({repr(self.operands)})"

    def __eq__(self, other):
        """Compare with another object."""
        return isinstance(other, type(self)) and self.operands == other.operands

    def __hash__(self) -> int:
        """Compute the hash of the object."""
        return hash((type(self), self.operands))


class Metric(Atomic):
    """A class for the metric function in PDDL."""

    MINIMIZE = Symbols.MINIMIZE.value
    MAXIMIZE = Symbols.MAXIMIZE.value

    def __init__(self, expression: FunctionExpression, optimization: str = MINIMIZE):
        """
        Initialize the metric.

        :param expression: functions to minimize or maximize.
        :param optimization: whether to minimize or maximize the function.
        """
        self._expression = expression
        self._optimization = optimization
        self._validate()

    @property
    def expression(self) -> FunctionExpression:
        """Get the functions."""
        return self._expression

    @property
    def optimization(self) -> str:
        """Get the optimization."""
        return self._optimization

    def instantiate(self, mapping: Mapping[Variable, Term]) -> "Metric":
        """Instantiate the metric with a mapping from variables to terms."""
        # metric should never be instantiated
        return self

    def _validate(self):
        """Validate the metric."""
        assert_(
            self.optimization in {self.MAXIMIZE, self.MINIMIZE},
            "Optimization metric not recognized.",
        )

    def __str__(self) -> str:
        """Get the string representation."""
        return f"{self.optimization} {self.expression}"

    def __repr__(self) -> str:
        """Get an unambiguous string representation."""
        return f"{type(self).__name__}({self.expression}, {self.optimization})"  # type: ignore

    def __eq__(self, other):
        """Override equal operator."""
        return (
            isinstance(other, Metric)
            and self.expression == other.expression
            and self.optimization == other.optimization
        )

    def __hash__(self):
        """Get the hash of a Metric."""
        return hash((self.expression, self.optimization))


class EqualTo(BinaryFunction, metaclass=BinaryOpMetaclass):
    """Equal to operator."""

    SYMBOL = Symbols.EQUAL


class LesserThan(BinaryFunction):
    """Lesser than operator."""

    SYMBOL = Symbols.LESSER


class LesserEqualThan(BinaryFunction):
    """Lesser or equal than operator."""

    SYMBOL = Symbols.LESSER_EQUAL


class GreaterThan(BinaryFunction):
    """Greater than operator."""

    SYMBOL = Symbols.GREATER


class GreaterEqualThan(BinaryFunction):
    """Greater or equal than operator."""

    SYMBOL = Symbols.GREATER_EQUAL


class Assign(BinaryFunction):
    """Assign operator."""

    SYMBOL = Symbols.ASSIGN

    def __init__(self, *operands: FunctionExpression):
        """
        Initialize the Assign operator.

        :param operands: the operands.
        """
        super().__init__(*operands)


class ScaleUp(BinaryFunction):
    """Scale-Up operator."""

    SYMBOL = Symbols.SCALE_UP

    def __init__(self, *operands: FunctionExpression):
        """
        Initialize the Scale-Up operator.

        :param operands: the operands.
        """
        super().__init__(*operands)


class ScaleDown(BinaryFunction):
    """Scale-Down operator."""

    SYMBOL = Symbols.SCALE_DOWN

    def __init__(self, *operands: FunctionExpression):
        """
        Initialize the Scale-Down operator.

        :param operands: the operands.
        """
        super().__init__(*operands)


class Increase(BinaryFunction):
    """Increase operator."""

    SYMBOL = Symbols.INCREASE

    def __init__(self, *operands: FunctionExpression):
        """
        Initialize the Increase operator.

        :param operands: the operands.
        """
        super().__init__(*operands)


class Decrease(BinaryFunction):
    """Decrease operator."""

    SYMBOL = Symbols.DECREASE

    def __init__(self, *operands: FunctionExpression):
        """
        Initialize the Decrease operator.

        :param operands: the operands.
        """
        super().__init__(*operands)


class Minus(BinaryFunction, metaclass=BinaryOpMetaclass):
    """Minus operator."""

    SYMBOL = Symbols.MINUS

    def __init__(self, *operands: FunctionExpression):
        """
        Initialize the Minus operator.

        :param operands: the operands.
        """
        super().__init__(*operands)


class Plus(BinaryFunction, metaclass=BinaryOpMetaclass):
    """Plus operator."""

    SYMBOL = Symbols.PLUS

    def __init__(self, *operands: FunctionExpression):
        """
        Initialize the Plus operator.

        :param operands: the operands.
        """
        super().__init__(*operands)


class Times(BinaryFunction, metaclass=BinaryOpMetaclass):
    """Times operator."""

    SYMBOL = Symbols.TIMES

    def __init__(self, *operands: FunctionExpression):
        """
        Initialize the Times operator.

        :param operands: the operands.
        """
        super().__init__(*operands)


class Divide(BinaryFunction, metaclass=BinaryOpMetaclass):
    """Divide operator."""

    SYMBOL = Symbols.DIVIDE

    def __init__(self, *operands: FunctionExpression):
        """
        Initialize the Divide operator.

        :param operands: the operands.
        """
        super().__init__(*operands)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains functions to simplify formulas creation."""

from typing import Collection, List, Optional

from pddl.custom_types import namelike
from pddl.helpers.base import ensure_set
from pddl.logic.terms import Constant, Variable


def variables(s: str, types: Optional[Collection[namelike]] = None) -> List[Variable]:
    """
    Return a list of terms.

    >>> variables("a b c", types=["type_1", "type_2"])
    [Variable(a), Variable(b), Variable(c)]

    :param s: a string with space-separated valid names.
    :param types: a list of types.
    :return: a list of variables.
    """
    types = ensure_set(types)
    return [Variable(x, types) for x in s.split()]


def constants(s: str, type_: Optional[namelike] = None) -> List[Constant]:
    """
    Return a list of constants.

    >>> constants("a b c")
    [Constant(a), Constant(b), Constant(c)]

    :param s: a string with space-separated valid names.
    :param type_: the type.
    :return: a list of constants.
    """
    return [Constant(x, type_) for x in s.split()]



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains tests for a PDDL plan."""

import pytest

from pddl import parse_domain
from pddl.action import Action
from pddl.core import Domain, Plan, Problem
from pddl.exceptions import PDDLValidationError
from pddl.logic.base import Not
from pddl.logic.helpers import constants, variables
from pddl.logic.predicates import Predicate
from tests.conftest import BLOCKSWORLD_FILES


class TestPlanEmpty:
    """Test the empty plan."""

    def setup_method(self):
        """Set up the tests."""
        self.plan = Plan(actions=[])

    def test_str(self):
        """Test the name getter."""
        assert str(self.plan) == ""

    def test_actions(self):
        """Test the actions getter."""
        assert self.plan.actions == []


def test_build_simple_plan():
    """Test a simple PDDL plan."""
    a, b, c = constants("a b c")
    a1_name = "action_1"
    a2_name = "action_2"
    plan = Plan(actions=[(a1_name, [a]), (a2_name, [b, c]), (a1_name, [c])])

    assert plan


def test_plan_equality():
    """Test plan equality."""
    a, b, c = constants("a b c")
    a1_name = "action_1"
    a2_name = "action_2"
    plan1 = Plan(actions=[(a1_name, [a]), (a2_name, [b, c]), (a1_name, [c])])
    plan2 = Plan(actions=[(a1_name, [a]), (a2_name, [b, c]), (a1_name, [c])])
    plan3 = Plan(actions=[(a1_name, [a]), (a2_name, [b, c])])

    assert plan1 == plan2
    assert plan1 != plan3


def test_bad_plan_action_name():
    """Test validator detects plan actions with invalid names."""
    a, b, c = constants("a b c")
    a1_name = "action_1"
    plan = Plan(actions=[(a1_name, [a, b]), (a1_name, [c, b])])

    x, y = variables("x y")
    p = Predicate("p", x, y)
    action_1 = Action("action_a", [x, y], precondition=p, effect=Not(p))
    domain = Domain("simple_domain", predicates={p}, actions={action_1})
    problem = Problem("simple_problem", domain, objects={a, b, c})

    with pytest.raises(
        PDDLValidationError,
        match="Action schema name of .* not found in domain.",
    ):
        plan.check(domain, problem)


def test_bad_plan_action_instantiation():
    """Test validator detects plan actions with invalid number of parameters."""
    a, b, c = constants("a b c")
    a1_name = "action_1"
    a2_name = "action_2"
    plan = Plan(actions=[(a1_name, [a]), (a2_name, [b, c, c]), (a1_name, [c, b, a])])

    x, y, z = variables("x y z")
    p = Predicate("p", x, y, z)
    action_1 = Action("action_1", [x, y, z], precondition=p, effect=Not(p))
    action_2 = Action("action_2", [x, y, z], precondition=Not(p), effect=p)
    domain = Domain("simple_domain", predicates={p}, actions={action_1, action_2})
    problem = Problem("simple_problem", domain, objects={a, b, c})

    with pytest.raises(
        PDDLValidationError,
        match="Number of parameters of .* does not match with the action schema in the domain.",
    ):
        plan.check(domain, problem)


def test_bad_plan_objects():
    """Test validator detects bad objects in plan."""
    a, b, c = constants("a b c")
    a1_name = "action_1"
    a2_name = "action_2"
    plan = Plan(actions=[(a1_name, [a, b, c]), (a2_name, [b, c, a])])

    a, b = constants("a b")
    x, y, z = variables("x y z")
    p = Predicate("p", x, y, z)
    action_1 = Action("action_1", [x, y, z], precondition=p, effect=Not(p))
    action_2 = Action("action_2", [x, y, z], precondition=Not(p), effect=p)
    domain = Domain(
        "simple_domain", constants={a}, predicates={p}, actions={action_1, action_2}
    )

    d, e, f = constants("d e f")
    problem = Problem("simple_problem", domain, objects={d, e, f})

    with pytest.raises(
        PDDLValidationError,
        match="Some parameters of .* are not in the problem objects.",
    ):
        plan.check(domain, problem)


def test_plan_instantiation_eq():
    """Test Issue #176."""
    domain = parse_domain(BLOCKSWORLD_FILES / "domain.pddl")
    b1, b2 = constants("badblock1 badblock2")
    plan = Plan(actions=[("pick-up", [b1, b2])])
    instantiated_plan = plan.instantiate(domain)
    assert len(instantiated_plan) == 1
    assert instantiated_plan[0].name == "pick-up"
    assert instantiated_plan[0].parameters == (b1, b2)
    assert "badblock" not in str(domain)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Formatting utilities for PDDL domains and problems."""

from typing import Callable, Collection, Dict, List, Optional, TypeVar

from pddl.custom_types import name
from pddl.logic.functions import NumericFunction
from pddl.logic.terms import Constant

T = TypeVar("T", name, NumericFunction)


def remove_empty_lines(s: str) -> str:
    """Remove empty lines from string."""
    return "\n".join(filter(str.strip, s.splitlines()))


def sort_and_print_collection(
    prefix,
    collection: Collection,
    postfix,
    to_string: Callable = str,
    is_mandatory: bool = False,
):
    r"""Produce the string of a PDDL section for a collection (e.g., requirements, actions, objects, etc.).

    Prefix starts the PDDL section, like "(:requirements"  or "(:actions"
    Postfix ends the section, usually with ")\n"
    The collection is sorted and printed as a string, using to_string to convert each element to a string.

    Args:
        prefix (str): start of the string
        collection (Collection): the collection of entities to report as a string
        postfix (str): the end of the string
        to_string (Callable, optional): the function to use to convert to string. Defaults to str.
        is_mandatory (bool, optional): if the string is mandatory even if the collection is empty. Defaults to False.

    Returns:
        str: a string with <prefix> <string of collection> <postfix>
    """
    if len(collection) > 0:
        return prefix + " ".join(sorted(map(to_string, collection))) + postfix
    elif is_mandatory:
        return prefix + postfix
    return ""


def print_types_or_functions_with_parents(
    prefix: str,
    types_dict: Dict[T, Optional[name]],
    postfix: str,
    to_string: Callable = str,
):
    """Print the type dictionary of a PDDL domain."""
    name_by_obj: Dict[Optional[T], List[name]] = {}
    for obj_name, parent_type in types_dict.items():
        name_by_obj.setdefault(parent_type, []).append(obj_name)  # type: ignore
    if not bool(name_by_obj):
        return ""
    return print_typed_lists(prefix, name_by_obj, postfix, to_string)


def print_constants(
    prefix, constants: Collection[Constant], postfix, to_string: Callable = str
):
    """Print constants in a PDDL domain."""
    term_by_type_tags: Dict[Optional[name], List[name]] = {}
    for c in constants:
        term_by_type_tags.setdefault(c.type_tag, []).append(c.name)
    if not bool(term_by_type_tags):
        return ""
    return print_typed_lists(prefix, term_by_type_tags, postfix, to_string)


def print_predicates_with_types(predicates: Collection):
    """Generate a string with predicates with type tags for the :predicates section."""
    result = ""
    for p in sorted(predicates):
        if p.arity == 0:
            result += f"({p.name})"
        else:
            result += f"({p.name}"
            for t in p.terms:
                if len(t.type_tags) > 1:
                    result += f" ?{t.name} - (either {' '.join(sorted(t.type_tags))})"
                else:
                    result += (
                        f" ?{t.name} - {sorted(t.type_tags)[0]}"
                        if t.type_tags
                        else f" ?{t.name}"
                    )
            result += ") "
        result += " "
    return result.strip()


def print_function_skeleton(function: NumericFunction) -> str:
    """Callable to print a function skeleton with type tags."""
    result = ""
    if function.arity == 0:
        result += f"({function.name})"
    else:
        result += f"({function.name}"
        for t in function.terms:
            result += (
                f" ?{t.name} - {sorted(t.type_tags)[0]}"
                if t.type_tags
                else f" ?{t.name}"
            )
        result += ")"
    return result


def print_typed_lists(
    prefix,
    names_by_obj: Dict[Optional[T], List[name]],
    postfix,
    to_string: Callable = str,
):
    """Print typed lists."""
    result = prefix

    if ":types" in prefix:
        # for mypy, convert names_by_obj to a dict with keys being strings (or None) and values list of strings
        super_to_subs = {
            str(k) if k is not None else None: v for k, v in names_by_obj.items()
        }

        # first print those types that have no parents
        subtypes = set()
        for subs in super_to_subs.values():
            subtypes |= set(subs)

        no_parents = set(super_to_subs.keys()) - subtypes
        no_parents |= set(super_to_subs.get(None, []))
        no_parents |= set(super_to_subs.get("object", []))
        no_parents.discard(None)
        no_parents.discard("object")

        result += (
            "\n    "
            + " ".join(sorted(to_string(n) for n in no_parents))
            + " - object\n"
        )

        # sort the rest of the parent types based on (1) if their subtypes are already defined and (2) their name
        remaining = sorted(str(k) for k in super_to_subs if k not in ("object", None))
        ind = 0
        while remaining:
            # if the index is out of range, we have a cycle in the types
            if ind >= len(remaining):
                raise ValueError("Cycle in types")

            # get the first type and its subtypes
            parent = remaining[ind]

            # check if parent is a subtype of something remaining
            allsubtypes = set()
            for k in remaining:
                allsubtypes |= set(super_to_subs[k])
            if parent in allsubtypes:
                ind += 1
            else:
                # print the parent and its subtypes
                result += "    " + (
                    " ".join(sorted(to_string(n) for n in super_to_subs[parent]))
                    + " - "
                    + to_string(parent)
                    + "\n"
                )
                # remove the parent from the remaining list
                remaining.remove(parent)
                # reset the index to 0
                ind = 0

    else:

        result += " "

        # names with no type will be printed at the end
        names_with_none_types = names_by_obj.pop(None, [])

        # print typed constants, first sorted by type, then by constant name
        for type_tag, typed_names in sorted(
            names_by_obj.items(), key=lambda type_and_name: type_and_name[0]  # type: ignore
        ):
            result += (
                " ".join(sorted(to_string(n) for n in typed_names)) + " - " + type_tag + " "  # type: ignore
            )

        if len(names_with_none_types) == 0:
            return result.strip() + postfix

        # print constants with no type
        result += " ".join(sorted(to_string(n) for n in names_with_none_types))

        if result == prefix + " ":
            result = result[:-1]

    result += postfix

    return result


def domain_to_string(domain) -> str:
    """Print a PDDL domain object."""
    return str(domain)


def problem_to_string(problem) -> str:
    """Print a PDDL problem object."""
    return str(problem)


def plan_to_string(plan) -> str:
    """Print a PDDL plan object."""
    return str(plan)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains tests for a PDDL domain."""

import copy
import pickle  # nosec
import re

import pytest

from pddl.action import Action
from pddl.core import Domain
from pddl.exceptions import PDDLValidationError
from pddl.logic import Constant, Variable
from pddl.logic.base import And, Not
from pddl.logic.functions import Decrease
from pddl.logic.functions import EqualTo as FunctionEqualTo
from pddl.logic.functions import (
    GreaterEqualThan,
    GreaterThan,
    Increase,
    LesserEqualThan,
    LesserThan,
    NumericFunction,
    NumericValue,
)
from pddl.logic.helpers import constants, variables
from pddl.logic.predicates import DerivedPredicate, Predicate
from pddl.parser.symbols import Symbols
from pddl.requirements import Requirements
from tests.conftest import pddl_objects_domains


@pytest.mark.parametrize("domain_obj", pddl_objects_domains)
def test_pickle_domain(domain_obj: Domain) -> None:
    """Test that domain objects can be pickled correctly."""
    domain_obj_bytes = pickle.dumps(domain_obj)  # nosec
    actual_domain_obj = pickle.loads(domain_obj_bytes)  # nosec
    assert domain_obj == actual_domain_obj


@pytest.mark.parametrize("domain_obj", pddl_objects_domains)
def test_deepcopy_domain(domain_obj: Domain) -> None:
    """Test that domain objects can be deepcopied correctly."""
    new_domain_obj = copy.deepcopy(domain_obj)
    assert domain_obj == new_domain_obj


class TestDomainEmpty:
    """Test the empty domain."""

    def setup_method(self):
        """Set up the tests."""
        self.domain = Domain("empty_domain")

    def test_name(self):
        """Test the name getter."""
        assert self.domain.name == "empty_domain"

    def test_requirements(self):
        """Test the requirements getter."""
        assert self.domain.requirements == set()

    def test_constants(self):
        """Test the constants getter."""
        assert self.domain.constants == set()

    def test_predicates(self):
        """Test the predicates getter."""
        assert self.domain.predicates == set()

    def test_actions(self):
        """Test the actions getter."""
        assert self.domain.actions == set()


def test_build_simple_domain():
    """Test a simple PDDL domain."""
    a, b, c = constants("a b c")
    x, y, z = variables("x y z")
    p = Predicate("p", x, y, z)
    action_1 = Action("action_1", [x, y, z], precondition=p, effect=Not(p))
    domain = Domain(
        "simple_domain", constants={a, b, c}, predicates={p}, actions={action_1}
    )

    assert domain


def test_build_simple_domain_with_derived_predicates():
    """Test a simple PDDL domain with derived predicates."""
    x, y, z = variables("x y z")
    p = Predicate("p", x, y, z)
    q = Predicate("q")
    r = Predicate("r")
    dp = DerivedPredicate(p, And(q, r))
    action_1 = Action("action_1", [x, y, z], precondition=p, effect=Not(p))
    domain = Domain(
        "simple_domain",
        predicates={p},
        derived_predicates={dp},
        actions={action_1},
    )
    assert domain


def test_build_domain_with_numeric_fluents():
    """Test a PDDL domain with simple numeric fluents."""
    x, y, z = variables("x y z")
    p = Predicate("p", x, y, z)
    q = Predicate("q")
    r = Predicate("r")
    func1 = NumericFunction("f1", x, y)
    func2 = NumericFunction("f2")
    func3 = NumericFunction("f3")
    action_1 = Action(
        "action_1",
        [x, y, z],
        precondition=p
        & FunctionEqualTo(func1, NumericValue(0))
        & GreaterThan(func2, NumericValue(1))
        & LesserThan(func3, NumericValue(5)),
        effect=Not(p) | q,
    )
    action_2 = Action(
        "action_2",
        [x, y, z],
        precondition=r
        & GreaterEqualThan(func1, NumericValue(1))
        & LesserEqualThan(func2, NumericValue(5)),
        effect=Not(p) | q,
    )
    domain = Domain(
        "domain_with_numeric",
        requirements={Requirements.NUMERIC_FLUENTS},
        predicates={p},
        functions={func1: None, func2: None, func3: None},
        actions={action_1, action_2},
    )
    assert domain


def test_build_domain_with_action_cost():
    """Test a PDDL domain with action costs."""
    x, y, z = variables("x y z")
    p = Predicate("p", x, y, z)
    q = Predicate("q")
    r = Predicate("r")
    cost1 = NumericFunction("cost1", x, y)
    cost2 = NumericFunction("cost2")
    action_1 = Action(
        "action_1",
        [x, y, z],
        precondition=p
        & FunctionEqualTo(cost1, NumericValue(0))
        & GreaterThan(cost2, NumericValue(1)),
        effect=Not(p) & Increase(cost1, NumericValue(1)),
    )
    action_2 = Action(
        "action_2",
        [x, y, z],
        precondition=r & GreaterEqualThan(cost1, NumericValue(1)),
        effect=(Not(p) | q) & Decrease(cost2, NumericValue(1)),
    )
    domain = Domain(
        "domain_with_numeric",
        requirements={Requirements.NUMERIC_FLUENTS},
        predicates={p},
        functions={cost1: None, cost2: None},
        actions={action_1, action_2},
    )
    assert domain


def test_build_domain_with_total_cost():
    """Test a PDDL domain with total costs."""
    x, y, z = variables("x y z")
    p = Predicate("p", x, y, z)
    q = Predicate("q")
    total_cost = NumericFunction(Symbols.TOTAL_COST.value)
    action_1 = Action(
        "action_1",
        [x, y, z],
        precondition=p & q,
        effect=Not(p) & Increase(total_cost, NumericValue(1)),
    )
    domain = Domain(
        "domain_with_total_cost",
        requirements={Requirements.ACTION_COSTS},
        predicates={p},
        functions={total_cost: None},
        actions={action_1},
    )
    assert domain


def test_cycles_in_type_defs_not_allowed() -> None:
    """Test that type defs with cycles are not allowed."""
    with pytest.raises(
        PDDLValidationError, match="cycle detected in the type hierarchy: A -> B -> C"
    ):
        Domain(
            "dummy",
            requirements={Requirements.TYPING},
            types={"A": "B", "B": "C", "C": "A"},
        )


def test_object_must_not_be_subtype() -> None:
    """Test that when the `object` type is used as subtype we raise error."""
    my_type = "my_type"
    type_set = {Symbols.OBJECT.value: my_type}

    with pytest.raises(
        PDDLValidationError,
        match=rf"object must not have supertypes, but got 'object' is a subtype of '{my_type}'",
    ):
        Domain("test", requirements={Requirements.TYPING}, types=type_set)  # type: ignore


def test_constants_type_not_available() -> None:
    """Test that when a type of a constant is not declared we raise error."""
    a = Constant("a", type_tag="t1")

    my_type = "my_type"
    type_set = {my_type: None}

    with pytest.raises(
        PDDLValidationError,
        match=rf"types \['t1'\] of term {re.escape(repr(a))} are not in available types {{'{my_type}'}}",
    ):
        Domain("test", requirements={Requirements.TYPING}, constants={a}, types=type_set)  # type: ignore


def test_predicate_variable_type_not_available() -> None:
    """Test that when a type of a predicate variable is not declared we raise error."""
    x = Variable("a", type_tags={"t1", "t2"})
    p = Predicate("p", x)

    my_type = "my_type"
    type_set = {my_type: None}

    with pytest.raises(
        PDDLValidationError,
        match=rf"types \['t1', 't2'\] of term {re.escape(repr(x))} are not in available types {{'{my_type}'}}",
    ):
        Domain("test", requirements={Requirements.TYPING}, predicates={p}, types=type_set)  # type: ignore


def test_action_parameter_type_not_available() -> None:
    """Test that when a type of a action parameter is not declared we raise error."""
    x = Variable("a", type_tags={"t1", "t2"})
    action = Action("p", [x])

    my_type = "my_type"
    type_set = {my_type: None}

    with pytest.raises(
        PDDLValidationError,
        match=rf"types \['t1', 't2'\] of term {re.escape(repr(x))} are not in available types {{'{my_type}'}}",
    ):
        Domain("test", requirements={Requirements.TYPING}, actions={action}, types=type_set)  # type: ignore


def test_derived_predicate_type_not_available() -> None:
    """Test that when a type of a term of a derived predicate is not declared we raise error."""
    x = Variable("a", type_tags={"t1", "t2"})
    p = Predicate("p", x)
    dp = DerivedPredicate(p, And())

    my_type = "my_type"
    type_set = {my_type: None}

    with pytest.raises(
        PDDLValidationError,
        match=rf"type '(t1|t2)' of term {re.escape(repr(x))} in atomic expression {re.escape(repr(p))} is not in "
        f"available types {{'{my_type}'}}",
    ):
        Domain("test", requirements={Requirements.TYPING}, derived_predicates={dp}, types=type_set)  # type: ignore



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""The setup script."""

import glob
import os

from setuptools import find_packages, setup

with open("README.md") as readme_file:
    readme = readme_file.read()

here = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(here, "pddl", "__version__.py"), "r") as f:
    exec(f.read(), about)


install_requires = ["lark>=1.1.5,<1.2.0", "click>=8.1.3,<9.0.0"]

setup(
    name=about["__title__"],
    description=about["__description__"],
    version=about["__version__"],
    author=about["__author__"],
    url=about["__url__"],
    author_email=about["__author_email__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Education",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
    ],
    python_requires=">=3.10",
    install_requires=install_requires,
    license=about["__license__"],
    include_package_data=True,
    data_files=[
        ("pddl/parser", glob.glob("pddl/parser/*.lark")),
    ],
    keywords="pddl",
    packages=find_packages(include=["pddl*"]),
    entry_points={
        "console_scripts": ["pddl=pddl.__main__:cli"],
    },
    test_suite="tests",
    tests_require=["pytest"],
    zip_safe=False,
)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains the tests for the domain parser."""

from textwrap import dedent

import pytest

from pddl.exceptions import PDDLParsingError
from pddl.logic.base import And
from pddl.logic.functions import (
    EqualTo,
    GreaterEqualThan,
    GreaterThan,
    NumericFunction,
    NumericValue,
)
from pddl.logic.terms import Constant
from pddl.parser.problem import ProblemParser
from pddl.requirements import Requirements


def test_problem_requirements_section_parsed() -> None:
    """Check that the requirements section is parsed correctly."""
    problem_str = dedent("""
        (define (problem test-problem)
            (:domain test-domain)
            (:requirements :typing)
            (:objects a b c)
            (:init (p b b b))
            (:goal (g a a a))
        )""")
    problem = ProblemParser()(problem_str)

    assert problem.requirements == {Requirements.TYPING}


def test_problem_objects_repetition_in_simple_typed_lists_not_allowed() -> None:
    """Check objects repetition in simple typed lists is detected and a parsing error is raised."""
    problem_str = dedent("""
    (define (problem test-problem)
        (:domain test-domain)
        (:requirements :typing)
        (:objects a b c a)
        (:init )
        (:goal (and ))
    )
    """)

    with pytest.raises(
        PDDLParsingError,
        match=".*error while parsing tokens \\['a', 'b', 'c', 'a'\\]: "
        "duplicate name 'a' in typed list already present",
    ):
        ProblemParser()(problem_str)


def test_problem_objects_repetition_in_typed_lists_not_allowed() -> None:
    """Check objects repetition in typed lists is detected and a parsing error is raised."""
    problem_str = dedent("""
    (define (problem test-problem)
        (:domain test-domain)
        (:requirements :typing)
        (:objects a - t1 b - t2 c - t3 a - t4)
        (:init )
        (:goal (and ))
    )
    """)

    with pytest.raises(
        PDDLParsingError,
        match=r".*error while parsing tokens \['a', '-', 't1', 'b', '-', 't2', 'c', '-', 't3', 'a', '-', 't4'\]: "
        r"duplicate name 'a' in typed list already inherits from types \['t1'\]",
    ):
        ProblemParser()(problem_str)


def test_problem_init_predicate_arg_repetition_allowed() -> None:
    """Check argument repetition in predicate list is allowed."""
    problem_str = dedent("""
    (define (problem test-problem)
        (:domain test-domain)
        (:requirements :typing)
        (:objects a)
        (:init (p a a))
        (:goal (and ))
    )
    """)
    ProblemParser()(problem_str)


def test_problem_init_predicate_repetition_name_allowed() -> None:
    """Check predicate repetition in init condition is allowed."""
    problem_str = dedent("""
    (define (problem test-problem)
        (:domain test-domain)
        (:requirements :typing)
        (:objects a)
        (:init (p a a) (p a a))
        (:goal (and ))
    )
    """)
    ProblemParser()(problem_str)


def test_numeric_comparison_in_goal() -> None:
    """Try to parse a goal with a numeric condition."""
    problem_str = dedent("""
    (define (problem hello-3-times)
        (:domain hello-world-functions)

        (:init
            ; if this was undefined, some planners would not assumed `0`
            (= (hello_counter jimmy) 0)
        )

        (:goal
            (> 5 3)
        )
    )
    """)
    problem = ProblemParser()(problem_str)
    assert problem.init == {
        EqualTo(NumericFunction("hello_counter", Constant("jimmy")), NumericValue(0))
    }
    assert problem.goal == GreaterThan(NumericValue(5), NumericValue(3))


def test_numeric_function_comparison_in_goal() -> None:
    """Try to parse a goal with a numeric condition and function."""
    problem_str = dedent("""
    (define (problem hello-3-times)
        (:domain hello-world-functions)

        (:init
            ; if this was undefined, some planners would not assumed `0`
            (= (hello_counter jimmy) 0)
        )

        (:goal
            (>= (hello_counter jimmy) 3)
        )
    )
    """)
    problem = ProblemParser()(problem_str)
    assert problem.init == {
        EqualTo(NumericFunction("hello_counter", Constant("jimmy")), NumericValue(0))
    }
    assert problem.goal == GreaterEqualThan(
        NumericFunction("hello_counter", Constant("jimmy")), NumericValue(3)
    )


def test_numeric_function_equality_in_goal() -> None:
    """Try to parse a goal with a numeric condition and function."""
    problem_str = dedent("""
    (define (problem hello-3-times)
        (:domain hello-world-functions)

        (:init
            ; if this was undefined, some planners would not assumed `0`
            (= (hello_counter jimmy) 0)
            (= (hello_counter jammy) 0)
        )

        (:goal
            (and
                (= 3 (hello_counter jimmy))
                (= (hello_counter jammy) 5)
            )
        )
    )
    """)
    problem = ProblemParser()(problem_str)
    assert problem.init == {
        EqualTo(NumericFunction("hello_counter", Constant("jimmy")), NumericValue(0)),
        EqualTo(NumericFunction("hello_counter", Constant("jammy")), NumericValue(0)),
    }
    assert problem.goal == And(
        EqualTo(NumericValue(3), NumericFunction("hello_counter", Constant("jimmy"))),
        EqualTo(NumericFunction("hello_counter", Constant("jammy")), NumericValue(5)),
    )



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#
"""
Core module of the package.

It contains the class definitions to build and modify PDDL domains or problems.
"""

from textwrap import indent
from typing import AbstractSet, Collection, Dict, List, Optional, Tuple, cast

from pddl._validation import (
    Functions,
    TypeChecker,
    Types,
    _check_types_in_has_terms_objects,
    validate,
)
from pddl.action import Action
from pddl.custom_types import name as name_type
from pddl.custom_types import namelike, parse_name, to_names, to_types  # noqa: F401
from pddl.formatter import (
    print_constants,
    print_function_skeleton,
    print_predicates_with_types,
    print_types_or_functions_with_parents,
    remove_empty_lines,
    sort_and_print_collection,
)
from pddl.helpers.base import assert_, check, ensure, ensure_set
from pddl.logic.base import And, Formula, is_literal
from pddl.logic.functions import FunctionExpression, Metric, NumericFunction
from pddl.logic.predicates import DerivedPredicate, Predicate
from pddl.logic.terms import Constant
from pddl.parser.symbols import Symbols
from pddl.requirements import Requirements


class Domain:
    """A class for a PDDL domain file."""

    def __init__(
        self,
        name: namelike,
        requirements: Optional[Collection["Requirements"]] = None,
        types: Optional[Dict[namelike, Optional[namelike]]] = None,
        constants: Optional[Collection[Constant]] = None,
        predicates: Optional[Collection[Predicate]] = None,  # TODO cannot be empty
        derived_predicates: Optional[
            Collection[DerivedPredicate]
        ] = None,  # TODO cannot be empty
        functions: Optional[Collection[FunctionExpression]] = None,
        actions: Optional[Collection["Action"]] = None,
    ):
        """
        Initialize a PDDL domain.

        :param name: the name of the domain.
        :param requirements: the requirements supported.
        :param types: the hierarchy of supported types.
            types is a dictionary mapping a type name to its ancestor.
        :param constants: the constants.
        :param predicates: the predicates.
        :param functions: the functions.
            functions is a dictionary mapping a function to its type.
        :param derived_predicates: the derived predicates.
        :param actions: the actions.
        """
        self._name = parse_name(name)
        self._requirements = ensure_set(requirements)
        self._types = Types(types, self._requirements)
        self._constants = ensure_set(constants)
        self._predicates = ensure_set(predicates)
        self._derived_predicates = ensure_set(derived_predicates)
        self._actions = ensure_set(actions)
        self._functions = Functions(functions, self._requirements)

        self._check_consistency()

    def _check_consistency(self) -> None:
        """Check consistency of a domain instance object."""
        type_checker = TypeChecker(self._types, self.requirements)
        type_checker.check_type(self._constants)
        type_checker.check_type(self._predicates)
        type_checker.check_type(self._actions)
        _check_types_in_has_terms_objects(self._actions, self._types.all_types)  # type: ignore
        self._check_types_in_derived_predicates()

    def _check_types_in_derived_predicates(self) -> None:
        """Check types in derived predicates."""
        dp_list = (
            [dp.predicate for dp in self._derived_predicates]
            if self._derived_predicates
            else set()
        )
        _check_types_in_has_terms_objects(dp_list, self._types.all_types)

    @property
    def name(self) -> name_type:
        """Get the name."""
        return self._name

    @property
    def requirements(self) -> AbstractSet["Requirements"]:
        """Get the PDDL requirements for this domain."""
        return self._requirements

    @property
    def constants(self) -> AbstractSet[Constant]:
        """Get the constants."""
        return self._constants

    @property
    def predicates(self) -> AbstractSet[Predicate]:
        """Get the predicates."""
        return self._predicates

    @property
    def functions(self) -> Dict[NumericFunction, Optional[name_type]]:
        """Get the functions."""
        return self._functions.raw

    @property
    def derived_predicates(self) -> AbstractSet[DerivedPredicate]:
        """Get the derived predicates."""
        return self._derived_predicates

    @property
    def actions(self) -> AbstractSet["Action"]:
        """Get the actions."""
        return self._actions

    @property
    def types(self) -> Dict[name_type, Optional[name_type]]:
        """Get the type definitions, if defined. Else, raise error."""
        return self._types.raw

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, Domain)
            and self.name == other.name
            and self.requirements == other.requirements
            and self.types == other.types
            and self.constants == other.constants
            and self.predicates == other.predicates
            and self.functions == other.functions
            and self.derived_predicates == other.derived_predicates
            and self.actions == other.actions
        )

    def __str__(self) -> str:
        """Print a PDDL domain object."""
        result = f"(define (domain {self.name})"
        body = ""
        indentation = " " * 4
        body += sort_and_print_collection("(:requirements ", self.requirements, ")\n")
        body += print_types_or_functions_with_parents("(:types", self.types, ")\n")
        body += print_constants("(:constants", self.constants, ")\n")
        if self.predicates:
            body += f"(:predicates {print_predicates_with_types(self.predicates)})\n"
        if self.functions:
            body += print_types_or_functions_with_parents(
                "(:functions", self.functions, ")\n", print_function_skeleton
            )
        body += sort_and_print_collection(
            "",
            self.derived_predicates,
            "",
            to_string=lambda obj: f"({Symbols.DERIVED.value} "
            f"{print_predicates_with_types([obj.predicate])} "
            f"{obj.condition})\n",
        )
        body += sort_and_print_collection(
            "",
            self.actions,
            "",
            to_string=lambda obj: str(obj) + "\n",
        )
        result = result + "\n" + indent(body, indentation) + "\n)"
        result = remove_empty_lines(result)
        return result


class Problem:
    """A class for a PDDL problem file."""

    def __init__(
        self,
        name: namelike,
        domain: Optional[Domain] = None,
        domain_name: Optional[namelike] = None,
        requirements: Optional[Collection["Requirements"]] = None,
        objects: Optional[Collection["Constant"]] = None,
        init: Optional[Collection[Formula]] = None,
        goal: Optional[Formula] = None,
        metric: Optional[Metric] = None,
    ):
        """
        Initialize the PDDL problem.

        :param name: the name of the PDDL problem.
        :param domain: the PDDL domain.
        :param domain_name: the domain name. Must match with the domain object.
        :param requirements: the set of PDDL requirements.
        :param objects: the set of objects.
        :param init: the initial condition.
        :param goal: the goal condition.
        :param metric: the metric.
        """
        self._name = parse_name(name)
        self._domain: Optional[Domain]
        self._domain_name: name_type
        self._domain, self._domain_name = self._parse_domain_and_domain_name(
            domain, domain_name
        )
        self._requirements: Optional[AbstractSet[Requirements]] = (
            self._parse_requirements(domain, requirements)
        )
        self._objects: AbstractSet[Constant] = ensure_set(objects)
        self._init: AbstractSet[Formula] = ensure_set(init)
        self._goal: Formula = ensure(goal, And())
        self._metric: Optional[Metric] = metric
        validate(
            all(map(is_literal, self.init)),
            "Not all formulas of initial condition are literals!",
        )

        self._check_consistency()

    def _parse_domain_and_domain_name(
        self,
        domain: Optional[Domain],
        domain_name: Optional[namelike],
    ) -> Tuple[Optional[Domain], name_type]:
        """
        Parse the domain and domain name.

        If the domain is given, use it. Otherwise, take the domain from the domain name.
        """
        if domain is not None and domain_name is not None:
            check(
                domain.name == domain_name,
                f"got both domain and domain_name, but domain_name differs: {domain.name} != {domain_name}",
                exception_cls=ValueError,
            )
            return domain, parse_name(domain_name)
        if domain is not None:
            return domain, domain.name
        if domain_name is not None:
            return None, parse_name(domain_name)
        raise ValueError("Either domain or domain_name must be given.")

    def _parse_requirements(
        self,
        domain: Optional[Domain],
        requirements: Optional[Collection[Requirements]],
    ) -> Optional[AbstractSet[Requirements]]:
        """
        Parse the requirements.

        If the requirements set is given, use it. Otherwise, take the requirements from the domain.
        """
        if requirements is None:
            return None if domain is None else domain.requirements

        # requirements is not None
        if domain is None:
            return set(cast(Collection[Requirements], requirements))

        check(
            ensure_set(requirements) == domain.requirements,
            f"got both requirements and domain, but requirements differ: {requirements} != {domain.requirements}",
            exception_cls=ValueError,
        )
        return domain.requirements

    def _check_consistency(self) -> None:
        """Check consistency of the PDDL Problem instance object."""
        if self._domain is not None:
            self.check(self._domain)

    def check(self, domain: Domain) -> None:
        """Check the problem definition against a domain definition."""
        validate(
            self.domain_name.lower() == domain.name.lower(),
            "Domain names don't match.",
        )
        validate(
            self._requirements is None or self._requirements == domain.requirements,
            "Requirements don't match.",
        )
        types = Types(domain.types, domain.requirements, skip_checks=True)  # type: ignore
        type_checker = TypeChecker(types, domain.requirements)
        type_checker.check_type(self.objects)
        type_checker.check_type(self.init)
        type_checker.check_type(self.goal)

    @property
    def name(self) -> name_type:
        """Get the name."""
        return self._name

    @property
    def domain(self) -> Domain:
        """Get the domain."""
        assert_(self._domain is not None, "Domain is not set.")
        return cast(Domain, self._domain)

    @domain.setter
    def domain(self, domain: Domain) -> None:
        """Set the domain."""
        self._domain_name = domain.name
        self._domain = domain
        self._check_consistency()

    @property
    def domain_name(self) -> name_type:
        """Get the domain name."""
        if self._domain is not None:
            return self._domain.name

        assert_(self._domain_name is not None, "Domain name is not set.")
        return cast(name_type, self._domain_name)

    @property
    def requirements(self) -> AbstractSet["Requirements"]:
        """Get the requirements."""
        if self._domain is not None:
            return self._domain.requirements

        return self._requirements if self._requirements is not None else set()

    @property
    def objects(self) -> AbstractSet["Constant"]:
        """Get the set of objects."""
        return self._objects

    @property
    def init(self) -> AbstractSet[Formula]:
        """Get the initial state."""
        return self._init

    @property
    def goal(self) -> Formula:
        """Get the goal."""
        return self._goal

    @property
    def metric(self) -> Optional[Metric]:
        """Get the metric."""
        return self._metric

    def __eq__(self, other):
        """Compare with another object."""
        return (
            isinstance(other, Problem)
            and self.name == other.name
            and self._domain == other._domain
            and self.domain_name == other.domain_name
            and self.requirements == other.requirements
            and self.objects == other.objects
            and self.init == other.init
            and self.goal == other.goal
            and self.metric == other.metric
        )

    def __str__(self) -> str:
        """Print a PDDL problem object."""
        result = f"(define (problem {self.name})"
        body = f"(:domain {self.domain_name})\n"
        indentation = " " * 4
        body += sort_and_print_collection("(:requirements ", self.requirements, ")\n")
        if self.objects:
            body += print_constants("(:objects", self.objects, ")\n")
        body += sort_and_print_collection(
            "(:init ", self.init, ")\n", is_mandatory=True
        )
        body += f"{'(:goal ' + str(self.goal) + ')'}\n"
        body += f"{'(:metric ' + str(self.metric) + ')'}\n" if self.metric else ""
        result = result + "\n" + indent(body, indentation) + "\n)"
        result = remove_empty_lines(result)
        return result


class Plan:
    """A class for a PDDL plan."""

    def __init__(self, actions: List[Tuple[namelike, List["Constant"]]]):
        """
        Initialize the PDDL plan.

        :param actions: the actions of the plan.
        """
        self._actions: list[Tuple[namelike, List["Constant"]]] = actions

    @property
    def actions(self) -> List[Tuple[namelike, List["Constant"]]]:
        """Get the actions."""
        return self._actions

    def instantiate(self, domain: Domain) -> List[Action]:
        """Instantiate the actions in the plan."""
        name_to_schema = {action.name: action for action in domain.actions}
        ret = []
        for action_name, parameters in self._actions:
            if action_name not in name_to_schema:
                raise ValueError(f"Action {action_name} not found in domain.")
            action_schema = name_to_schema[action_name]
            instantiated_action = action_schema.instantiate(parameters)
            ret.append(instantiated_action)
        return ret

    def check(self, domain: Domain, problem: Problem) -> None:
        """Check the plan against a domain and a problem."""
        name_to_schema = {action.name: action for action in domain.actions}
        domain_actions = {action.name for action in domain.actions}
        all_objects = problem.objects | domain.constants
        for action_name, parameters in self._actions:
            action = self._action_to_string(action_name, parameters)
            validate(
                action_name in domain_actions,
                f"Action schema name of {action} not found in domain.",
            )
            validate(
                len(parameters) == len(name_to_schema[action_name].parameters),
                f"Number of parameters of {action} does not match with the action schema in the domain.",
            )
            validate(
                all(param in all_objects for param in parameters),
                f"Some parameters of {action} are not in the problem objects.",
            )

    def _action_to_string(
        self, action_name: namelike, parameters: List["Constant"]
    ) -> str:
        """Convert an action to a string."""
        params = " ".join(str(param) for param in parameters)
        return f"({action_name} {params})"

    def __eq__(self, other):
        """Compare with another object."""
        return isinstance(other, Plan) and self._actions == other._actions

    def __str__(self) -> str:
        """Print a PDDL plan object."""
        actions = []
        for action_name, parameters in self._actions:
            actions.append(self._action_to_string(action_name, parameters))
        return "\n".join(actions)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""This module contains tests for a PDDL problem."""

import copy
import pickle  # nosec

import pytest

from pddl.core import Domain, Problem
from pddl.exceptions import PDDLValidationError
from pddl.logic.base import And, Not
from pddl.logic.functions import EqualTo as FunctionEqualTo
from pddl.logic.functions import (
    GreaterEqualThan,
    GreaterThan,
    LesserEqualThan,
    LesserThan,
    Metric,
    NumericFunction,
)
from pddl.logic.helpers import constants, variables
from pddl.logic.predicates import Predicate
from pddl.parser.symbols import Symbols
from tests.conftest import pddl_objects_problems


@pytest.mark.parametrize("problem_obj", pddl_objects_problems)
def test_pickle_problem(problem_obj: Problem) -> None:
    """Test that problem objects can be pickled correctly."""
    problem_obj_bytes = pickle.dumps(problem_obj)  # nosec
    actual_problem_obj = pickle.loads(problem_obj_bytes)  # nosec
    assert problem_obj == actual_problem_obj


@pytest.mark.parametrize("problem_obj", pddl_objects_problems)
def test_deepcopy_problem(problem_obj: Problem) -> None:
    """Test that problem objects can be deepcopied correctly."""
    new_problem_obj = copy.deepcopy(problem_obj)
    assert problem_obj == new_problem_obj


class TestProblemEmpty:
    """Test the empty problem."""

    def setup_method(self):
        """Set up the tests."""
        self.domain = Domain("empty_domain")
        self.problem = Problem("empty_problem", self.domain)

    def test_name(self):
        """Test the name getter."""
        assert self.problem.name == "empty_problem"

    def test_requirements(self):
        """Test the requirements getter."""
        assert self.problem.requirements == set()

    def test_objects(self):
        """Test the objects getter."""
        assert self.problem.objects == set()

    def test_init(self):
        """Test the init getter."""
        assert self.problem.init == set()

    def test_goal(self):
        """Test the goal getter."""
        assert self.problem.goal == And()


def test_build_simple_problem():
    """Test a simple PDDL problem."""
    x, y, z = variables("x y z")
    o1, o2, o3 = constants("o1 o2 o3")
    p = Predicate("p", x, y, z)
    q = Predicate("q", x, y, z)
    problem = Problem(
        "simple_problem",
        domain_name="simple_domain",
        objects=[o1, o2, o3],
        init={p, Not(q)},
        goal=p & q,
    )
    assert problem


def test_build_problem_with_metric():
    """Test a PDDL problem with metric."""
    x, y = variables("x y")
    o1, o2 = constants("o1 o2")
    p = Predicate("p", x, y)
    q = Predicate("q")
    total_cost = NumericFunction(Symbols.TOTAL_COST.value)
    problem = Problem(
        "simple_problem",
        domain_name="simple_domain",
        objects=[o1, o2],
        init={p, Not(q), FunctionEqualTo(total_cost, 0)},
        goal=p & q,
        metric=Metric([total_cost]),
    )
    assert problem


def test_build_problem_with_metric_list():
    """Test a PDDL problem with two functions and metric."""
    x, y = variables("x y")
    o1, o2 = constants("o1 o2")
    p = Predicate("p", x, y)
    q = Predicate("q")
    cost1 = NumericFunction("cost1", x, y)
    cost2 = NumericFunction("cost2")
    problem = Problem(
        "simple_problem",
        domain_name="simple_domain",
        objects=[o1, o2],
        init={p, Not(q), FunctionEqualTo(cost1, 0), FunctionEqualTo(cost2, 1)},
        goal=p & q & GreaterEqualThan(cost1, 3) & LesserEqualThan(cost2, 10),
        metric=Metric([cost1, cost2], Metric.MAXIMIZE),
    )
    assert problem


def test_build_problem_with_numeric_goal():
    """Test a PDDL problem with numeric fluents in goal."""
    x, y = variables("x y")
    o1, o2 = constants("o1 o2")
    p = Predicate("p", x, y)
    q = Predicate("q")
    cost1 = NumericFunction("cost1", x, y)
    cost2 = NumericFunction("cost2")
    problem = Problem(
        "simple_problem",
        domain_name="simple_domain",
        objects=[o1, o2],
        init={p, Not(q), FunctionEqualTo(cost1, 0), FunctionEqualTo(cost2, 10)},
        goal=p & q & GreaterThan(cost1, 3) & LesserThan(cost2, 10),
    )
    assert problem


def test_problem_check_domain_name_mismatch() -> None:
    """Test domain name check fails when names differ beyond case."""
    domain = Domain("simple_domain")
    problem = Problem("simple_problem", domain_name="SIMPLE_DOMAIN2")

    with pytest.raises(PDDLValidationError, match="Domain names don't match."):
        problem.check(domain)


def test_problem_check_domain_name_case_insensitive_match() -> None:
    """Test domain name check accepts names that only differ by case."""
    domain = Domain("simple_domain")
    problem = Problem("simple_problem", domain_name="SIMPLE_DOMAIN")

    problem.check(domain)



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Main entrypoint for the PDDL parser CLI tool."""

import os
import sys

import click

from pddl import parse_domain, parse_plan, parse_problem
from pddl.formatter import domain_to_string, plan_to_string, problem_to_string


@click.command()
@click.argument("domain_file", type=click.Path(exists=True, dir_okay=False))
@click.argument(
    "problem_file",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
)
@click.argument(
    "plan_file",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
)
@click.option("-q", "--quiet", is_flag=True, help="Don't print anything.")
def cli(domain_file, problem_file, plan_file, quiet):
    """The unquestionable parser for PDDL 3.1."""  # noqa
    if quiet:
        sys.stdout = open(os.devnull, "a")
    domain = parse_domain(domain_file)
    print(domain_to_string(domain))
    if problem_file is None:
        return
    problem = parse_problem(problem_file)
    problem.check(domain)
    print(problem_to_string(problem))
    if plan_file is None:
        return
    plan = parse_plan(plan_file)
    plan.check(domain, problem)
    print(plan_to_string(plan))


if __name__ == "__main__":
    cli()



#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Test introduction documentation page."""

import logging
from io import StringIO
from unittest.mock import patch

from tests.conftest import ROOT_DIRECTORY
from tests.test_docs.base import BaseTestMarkdownDocs, compile_and_exec


class TestIntroduction(BaseTestMarkdownDocs):
    """Test that the code snippet in the introduction are correct."""

    MD_FILE = ROOT_DIRECTORY / "README.md"

    @classmethod
    def setup_class(cls):
        """Set up the test."""
        super().setup_class()
        cls.python_code_blocks = cls.extract_code_blocks("python")
        cls.output_code_blocks = cls.extract_code_blocks("output")
        cls.locals = {}

    def test_python_and_output(self):
        """Test Python snipped and its output."""
        for python_code_block, output_code_block in zip(
            self.python_code_blocks, self.output_code_blocks
        ):
            logging.debug(f"Testing block: {python_code_block[:50]}...")
            with patch("sys.stdout", new_callable=StringIO) as mock_stdout:
                compile_and_exec(python_code_block, locals_dict=self.locals)
                actual_output = mock_stdout.getvalue()
                expected_output = output_code_block
                assert actual_output == expected_output
